/*
 *	als object for Max/MSP (Ambient Light Sensor)
 *
 *		Copyright(C) 2006-2007 Shigeyuki Hirai
 */

#include "ext.h"
#include <mach/mach.h> 
#include <IOKit/IOKitLib.h> 
#include <CoreFoundation/CoreFoundation.h> 

#define DEFAULT_INTERVAL 250.

enum {
	kGetSensorReadingID = 0, // getSensorReading(int *, int *)  
	kGetLEDBrightnessID = 1, // getLEDBrightness(int, int *)  
	kSetLEDBrightnessID = 2, // setLEDBrightness(int, int, int *)  
	kSetLEDFadeID = 3, // setLEDFade(int, int, int, int *)  
	// other firmware-related functions  
	// verifyFirmwareID = 4, 
	// verifyFirmware(int *)  
	// getFirmwareVersionID = 5, 
	// getFirmwareVersion(int *)  
	// other flashing-related functions  
	// ... 
}; 

typedef struct als
{
	t_object 		m_ob;
	void 			*m_lmuleftout;
	void			*m_lmurightout;
	void 			*m_clock;
	double			m_interval;
	long 			m_run;
	io_service_t	m_serviceObject;  
	io_connect_t	m_dataPort; 
} t_als;

void accessALS(t_als *x);
void als_float(t_als *x, double f);
void als_int(t_als *x, long n);
void als_tick(t_als *x);
void als_bang(t_als *x);
void als_start(t_als *x);
void als_stop(t_als *x);
void als_rate(t_als *x, long rate);
void als_free(t_als *x);
void *als_new(void);

void *als_class;

void accessALS(t_als *x)
{
	kern_return_t kr;  
	IOItemCount scalarInputCount = 0;  
	IOItemCount scalarOutputCount = 2;  
	SInt32 left = 0, right = 0;
	
	if (x->m_dataPort == NULL) {
		error("als: couldn't get dataPort");
		return;
	}
	kr = IOConnectMethodScalarIScalarO(x->m_dataPort, kGetSensorReadingID,  scalarInputCount, scalarOutputCount, &left, &right);
	if (kr == KERN_SUCCESS) {
//		post("als: (%8d, %8d)", left, right);
		outlet_float(x->m_lmurightout, (float)left);
		outlet_float(x->m_lmuleftout, (float)right);
	} else if (kr == kIOReturnBusy) {
		error("als: device busy [%d]", kr);  
	} else {
		error("als: couldn't get data from light sensors");
	}
	return;
}

void als_float(t_als *x, double f)
{
	als_int(x, (long)f);
}

void als_tick(t_als *x)
{
	accessALS(x);
	if (x->m_run) clock_fdelay(x->m_clock, x->m_interval);
}

void als_bang(t_als *x)
{
	accessALS(x);
}

void als_start(t_als *x)
{
	als_int(x, 1);
}

void als_stop(t_als *x)
{
	als_int(x, 0);
}

void als_int(t_als *x, long n)
{
	if (n != 0) {
		if (x->m_run == 0) {
			x->m_run = 1;
			clock_delay(x->m_clock, 0);
		}
	} else {
		x->m_run = 0;
	}
}

void als_rate(t_als *x, long rate)
{
	if (rate > 0) {
		x->m_interval = 1000./rate;
	}
}

int als_init(t_als *x)
{
	kern_return_t kr;  

	// Look up a registered IOService object whose class is AppleLMUController  
	x->m_serviceObject = IOServiceGetMatchingService(kIOMasterPortDefault, IOServiceMatching("AppleLMUController"));  
	if (!x->m_serviceObject) {
		error("als: couldn't find ambient light sensors by AppleLMUController");
		x->m_serviceObject = IOServiceGetMatchingService(kIOMasterPortDefault, IOServiceMatching("IOI2CDeviceLMU"));  
		if (!x->m_serviceObject) {
			error("als: couldn't find ambient light sensors by IOI2CDeviceLMU");
			return -1;
		} else {
			post("als: found ambient light sensors by IOI2CDeviceLMU");
		}
	} else {
		post("als: found ambient light sensors by AppleLMUController");
	}

	// Create a connection to the IOService object  
	kr = IOServiceOpen(x->m_serviceObject, mach_task_self(), 0, &x->m_dataPort);  
	IOObjectRelease(x->m_serviceObject);  
	if (kr != KERN_SUCCESS) {
		error("als: IOServiceOpen(): %d", kr);  
		return (kr);
	}

	return 0;
}

void als_free(t_als *x)
{
	freeobject(x->m_clock);
	x->m_clock = NULL;
}

void *als_new(void)
{
	t_als *x = (t_als *)newobject(als_class);
	x->m_lmurightout = floatout(x);
	x->m_lmuleftout = floatout(x);
	x->m_clock = clock_new(x, (method)als_tick);
	x->m_run = 0;
	x->m_interval = DEFAULT_INTERVAL;
	x->m_serviceObject = NULL;
	x->m_dataPort = 0;
	if (als_init(x)) {
//		freeobject(x->m_clock);
//		x->m_clock = NULL;
//		return NULL;
	}
	return x;
}

void main()
{
	setup((t_messlist **)&als_class, 
			(method)als_new, 
			(method)als_free, 
			(short)sizeof(t_als), 
			0L, 
			0);
	addint((method)als_int);
	addfloat((method)als_float);
	addbang((method)als_bang);
	addmess((method)als_start, "start", 0);
	addmess((method)als_stop, "stop", 0);
	addmess((method)als_rate, "rate", A_LONG, 0);
}
